#######################################################################
# #####    Dominik Benke, M.Sc. programed this app during his   ##### #
# #####      PhD thesis at University Bayreuth 2019-2024        ##### #
#######################################################################

# import own source code
import sourceCode_ellipseDetection.tab1_startPage_v6 as tab1_startPage
import sourceCode_ellipseDetection.tab2_Page_Area_detection_v8 as tab2_Page_Area_detection
import sourceCode_ellipseDetection.tab3_Page_Area_Selection_v10 as tab3_Page_Area_Selection
import sourceCode_ellipseDetection.tab4_Page_EllipsePicking_v10 as tab4_Page_EllipsePicking

# import packages
import matplotlib
from matplotlib.figure import Figure
from matplotlib.pyplot import rcParams
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk
from tkinter import ttk
from tkinter import filedialog
from skimage import feature
from skimage import color, img_as_ubyte, io
from skimage.filters import rank
from skimage.segmentation import watershed
from skimage.morphology import disk
import os
from scipy import ndimage as ndi
from skimage.measure import regionprops
import datetime

# global plot parameters
matplotlib.use('TkAgg')
LARGE_FONT = ('Verdana', 12)
GlobFontSize = 10
params = {
    'axes.labelsize': GlobFontSize,
    'font.size': GlobFontSize,
    'axes.titlesize': GlobFontSize,
    'font.family': 'Calibri',
    'legend.fontsize': GlobFontSize,
    'xtick.labelsize': GlobFontSize,
    'ytick.labelsize': GlobFontSize,
    'text.usetex': False,
    'figure.constrained_layout.use': True,
    'lines.markersize': 3.0,
    'boxplot.flierprops.markersize': 3
}
rcParams.update(params)

# app code
class EllipsoidDetectionApp(tk.Tk):
    version = 'Version 0.54'
    print('Py file location: ', os.getcwd())
    print('used py file path: ', __file__)
    # administration for variables
    progr_count = 1
    variable_1 = 1.0
    # For diagnosis it is possible to print more data on console
    console_print_on = False
    initialdirectory = 'C:/'

    # load dummypic file from folder Testdata
    dummydirectory = os.getcwd() + '\\testdata'
    print('The dummy images are located in: ' + str(dummydirectory))
    dummypic = dummydirectory + '\\190607_DBmL_41B_14.bmp'
    # other dummys in folder:
    # '\\220125DBmLR61_DB13_15%_quer_0%_pos1__014_LSM2D.jpg'
    # '\\190607_DBmL_41A_10.tif'

    ## dummys for variables interchange ##
    # create all varaibles and images that are necessary for interchange
    # in between the different tabs of the program
    image = color.rgb2gray(io.imread(dummypic)[:630])
    image_rgb = color.rgb2gray(io.imread(dummypic)[:630])
    original = color.rgb2gray(io.imread(dummypic)[:630])
    # imagesize and new_imagesize is array of (height, width)
    imagesize = image_rgb.shape
    new_imagesize = image_rgb.shape
    # show initial dummy pxl ratio
    pxl_ratio = 11.16 #21.1
    edges = feature.canny(image_rgb)
    gradient = rank.gradient(img_as_ubyte(edges), disk(3)) < 10
    markers = ndi.label(gradient)[0]
    labels = watershed(edges, markers)
    ParticleProp = regionprops(labels)
    particles_at_border = False
    # create an empty image with size of dummypic
    # np.zero needs dimension of height, width --> using new_imagesize in normal order
    image_ellipses = np.zeros([new_imagesize[0], new_imagesize[1], 3])
    image_numbers = image_ellipses
    ellipse_draw_list = []
    image_path = ''
    ParticleList = []
    final_ParticleList = []
    WorkingList = []
    type_threshold = ''
    filter_used_list = []
    final_polar_plot = Figure()
    analyse_orientations_img = Figure()
    # figure needs size in width, height -> turn array of img.shape --> new_imagesize[1], new_imagesize[0]
    finalfig_orgdim = Figure(figsize=(new_imagesize[1] / 100, new_imagesize[0] / 100))

    # initial parameters for sliders
    sigma = 0.9
    low_threshold = 0.40
    high_threshold = 200
    labelLines_parameter = 1

    # parameters for in frame image size
    img_x = 6.0
    img_y = 3.7
    finalfigure = Figure(figsize=(img_x, img_y))

    # Initialization of the program
    # arguments and kwargs = keyword arguments (dictionaries)
    def __init__(self, *args, **kwargs):
        ## prepare GUI ##
        tk.Tk.__init__(self, *args, **kwargs)
        tk.Tk.wm_title(self, '2D Ellipse Particle Detector')

        # make Window fit full screen
        # but for images in thesis: 1000,700 instead screenwidth and pad
        pad = 2
        self.geometry('{0}x{1}+0+0'.format(self.winfo_screenwidth() - pad, self.winfo_screenheight() - pad))
        self.update()
        # print in console if parameter above is set True
        if self.console_print_on: print(self.winfo_width())
        if self.console_print_on: print(self.winfo_height())

        # define figure size within the full screen
        dpi = 100
        self.img_x = round(((self.winfo_width() - 220) / dpi) / 2, 2)   # for fig III-17 use 3
        self.img_y = round(((self.winfo_height() * 0.85) / dpi) / 2, 2)   # for fig III-17 use 0.6
        if self.console_print_on: print(self.img_x)
        if self.console_print_on: print(self.img_y)

        ## Implementation of menu bar ##
        menubar = tk.Menu(self)
        # 'File' dropdown menu
        filemenu = tk.Menu(menubar, tearoff=0)
        filemenu.add_command(label='New', command=lambda: self.open_new(self.tabControl))
        filemenu.add_command(label='New with Evaluation File', command=lambda: [self.load_evaluation_parameter(), self.open_new(self.tabControl)])
        self.load_menu_point = filemenu.add_command(label='Load Evaluation Parameter', command=self.load_evaluation_parameter, state='normal')
        filemenu.add_command(label='Close', command=self.close_current)
        menubar.add_cascade(label='File', menu=filemenu)
        # 'Edit' dropdown menu
        self.editmenu = tk.Menu(menubar, tearoff=0)
        self.automated_crop = tk.BooleanVar()
        self.editmenu.add_checkbutton(label='Automated Line finder', variable=self.automated_crop)
        self.no_crop = tk.BooleanVar()
        self.editmenu.add_checkbutton(label='No crop', variable=self.no_crop)
        self.image_w_axes = tk.BooleanVar()
        self.image_w_axes.set(True)
        self.editmenu.add_checkbutton(label='Images with axes', variable=self.image_w_axes, )
        self.thresholding_var = tk.BooleanVar()
        self.thresholding_var.set(False)
        self.editmenu.add_checkbutton(label='Thresholding Image', variable=self.thresholding_var)
        self.save_angle_dist = tk.BooleanVar()
        self.save_angle_dist.set(True)
        self.editmenu.add_checkbutton(label='Save angle distribution, too', variable=self.save_angle_dist)
        self.editmenu.add_separator()
        self.consol_print = tk.BooleanVar()
        self.editmenu.add_checkbutton(label='Console print out on', variable=self.consol_print, command=self.toogle_consoleprint)
        self.editmenu.add_command(label='Credits', command=self.credits_win)
        menubar.add_cascade(label='Edit', menu=self.editmenu)
        # mount menubar on window
        self.menubar = menubar
        self.config(menu=self.menubar)

        # create dummy tabs
        self.new_tabcontrol()

        # start program
        self.new_start()

    # switch console print outs on or off by menu button
    def toogle_consoleprint(self):
        if self.consol_print.get():
            self.console_print_on = True
            print('console on')
        else:
            self.console_print_on = False
            print('console off')

    # create new dummy tabs
    def new_tabcontrol(self):
        # tab control frames
        tabControl = ttk.Notebook(self)
        tabControl.pack(side='top', fill='both', expand=True)
        # #tabID 0
        self.tab1 = ttk.Frame(tabControl)
        tabControl.add(self.tab1, text='Image')
        # #tabID 1
        self.tab2 = ttk.Frame(tabControl)
        tabControl.add(self.tab2, text='Edge Detection')
        # #tabID 2
        self.tab3 = ttk.Frame(tabControl)
        tabControl.add(self.tab3, text='Area Selection')
        # #tabID 3
        self.tab4 = ttk.Frame(tabControl)
        tabControl.add(self.tab4, text='Ellipse Picking')  # Statistics & Evaluation
        # #hide tabs 1-3, show tab 0
        tabControl.hide(1)
        tabControl.hide(2)
        tabControl.hide(3)
        tabControl.pack(side='top', fill='both', expand=True)
        self.tabControl = tabControl

    # start the evaluation of an image by loading the image with or without parameters in new popup frame
    def new_start(self):
        choose = tk.Tk()
        choose.geometry('250x110')
        choose.wm_title('Start up')
        choose_label1 = tk.Label(master=choose, text='Welcome to Dominik Benkes ellipse detection.', anchor='center')
        choose_label2 = tk.Label(master=choose, text=self.version)
        choose_label1.grid(row=0, sticky='EW')
        choose_label2.grid(row=1)
        choose_button1 = ttk.Button(master=choose, text='Load File', command=lambda: [self.next_step(1), choose.destroy()])
        choose_button1.grid(row=3)
        choose_button1 = ttk.Button(master=choose, text='Load with pre-set for evaluation', command=lambda: [self.load_evaluation_parameter(), self.next_step(1), choose.destroy()], state='enabled')
        choose_button1.grid(row=4)
        self.progr_count = 1

    # go to the next part/tab of the evaluation by clicking on the respective button in the tab 
    # the program counter increases
    def next_step(self, progr_count):
        self.progr_count = progr_count
        # bind content of next steps to frames/tabs
        if self.progr_count == 1:
            if self.console_print_on: print('Program counter: ' +str(self.progr_count))
            self.load_image()
            # north south east west = nsew    
            self.content_tab1 = tab1_startPage.StartPage(self.tab1, self)
            self.content_tab1.grid(row=0, column=0, sticky='nsew')
        elif self.progr_count == 2:
            if self.console_print_on: print('Program counter: ' +str(self.progr_count))
            self.content_tab2 = tab2_Page_Area_detection.Page_Area_detection(self.tab2, self)
            self.content_tab2.grid(row=0, column=0, sticky='nsew')
        elif self.progr_count == 3:
            if self.console_print_on: print('Program counter: ' +str(self.progr_count))
            self.content_tab3 = tab3_Page_Area_Selection.Page_Area_Selection(self.tab3, self)
            self.content_tab3.grid(row=0, column=0, sticky='nsew')
        elif self.progr_count == 4:
            if self.console_print_on: print('Program counter: ' +str(self.progr_count))
            self.content_tab4 = tab4_Page_EllipsePicking.Page_EllipsePicking(self.tab4, self)
            self.content_tab4.grid(row=0, column=0, sticky='nsew')
        else:
            None
        if self.console_print_on: print('Evaluations since program start: ' + str(self.variable_1))
        if self.console_print_on: print(self.tabControl.index('end'))

    # show the chosen tab on the window 
    def show_frame(self, cont):
        self.tabControl.select(cont)

    # close the current evaluation
    def close_current(self):
        variable_2 = self.variable_1+1
        print(variable_2)
        self.variable_1 = variable_2
        self.tabControl.destroy()    

    # start a new evaluation
    def open_new(self, container):
        self.new_dummys()
        # Refresh of tab control
        self.tabControl.destroy()    
        self.new_tabcontrol()
        self.progr_count = 1
        self.next_step(1)
        if self.console_print_on: print('open new file')

    # save the data from all tabs
    def write_datafile(self):
        '''
        Variables to be saved:
        file A:     self.final_ParticleList  //List

        file B:     self.version
                    self.progr_count
                    self.initialdirectory
                    self.pxl_ratio
                    self.thresholding_var
                    self.particles_at_border
                    self.sigma
                    self.low_threshold
                    self.high_threshold
                    self.labelLines_parameter
                    self.ellipse_draw_list   //List

        Img colored ellipses:       self.finalfigure
        Plot of orientations:       self.final_polar_plot
        '''
        self.save_window()
        #progress bar for saving
        self.progressbar.step(10)
        # save parameters of tab 1
        self.content_tab1.make_save_update_from_tab1(self)
        self.progressbar.step(10)
        # save parameters of tab 2, if present
        if self.progr_count>1:
            self.content_tab2.make_save_update_from_tab2(self)
        self.progressbar.step(10)
        # save parameters of tab 3, if present
        if self.progr_count>2:
            self.content_tab3.make_save_update_from_tab3(self)
        self.progressbar.step(10)
        # save parameters of tab 4, if present
        if self.progr_count>3:
            self.content_tab4.make_save_update_from_tab4(self)
        self.progressbar.step(10)

        ## write file A - Datafile ##
        # #create filename with timestamp
        time =  datetime.datetime.now().strftime('%Y-%m-%d %H-%M-%S')
        split_name = os.path.splitext(str(self.image_path[1]))
        new_name = str(self.image_path[0])+'\\'+str(split_name[0])+'_evaluation_'+time+'.txt'
        if self.console_print_on: print(new_name)
        # #write ellipse details to file        
        print_List = self.final_ParticleList
        if print_List!=[]:
            fileA =  open(str(new_name),'w')
            for i in print_List:
                # removes brakets on start and end 
                # + Add filename in front, just not on first row (column names)
                if str(i)[0] =='P':
                    j = str(self.file_name)+', '+str(i)
                else:
                    j = str(self.file_name)+', '+str(i)[1:-1]
                fileA.write('%s\n' % j)
            fileA.close()
            if self.console_print_on: print('Particle List saved')
        else:
            if self.console_print_on: print('No particle data to save')
        self.progressbar.step(25)

        ## save figures ##
        # figure with axes but with low white space around (bbox and pad_inches)
        self.finalfigure.savefig((str(self.image_path[0]) + '\\' + str(split_name[0]) + '_evaluation_' + time + '.png'), dpi=100, bbox_inches='tight', pad_inches=0)
        # figures polar plot and orientation analysis, if used at tab 4
        if self.save_angle_dist.get():
            self.final_polar_plot.savefig((str(self.image_path[0]) + '\\' + str(split_name[0]) + '_orientation_' + time + '.png'), dpi=155, bbox_inches='tight', pad_inches=0)
        if len(self.analyse_orientations_img.get_children()) > 1:
            self.analyse_orientations_img.savefig((str(self.image_path[0]) + '\\' + str(split_name[0]) + '_analysed_orientations_' + time + '.png'), dpi=300, bbox_inches='tight', pad_inches=0)
            print('analysed orientations' + str(len(self.analyse_orientations_img.get_children())))
        # save the final figure with overlay and the overlay only; both with original dimensions 
        # same scale (pxl/nm) as original image
        self.finalfig_orgdim.savefig((str(self.image_path[0]) + '\\' + str(split_name[0]) + '_evaluation2_' + time + '.png'), dpi=1)
        plt.imsave(fname=(str(self.image_path[0]) + '\\' + str(split_name[0]) + '_evaluation3_' + time + '.png'), arr=self.image_ellipses, format='png')

        ## write file B - config-data for re-evaluation ##
        config_file_name = str(self.image_path[0]) + '\\' + str(split_name[0]) + '_progr_data_' + time + '.txt'
        if self.console_print_on: print(config_file_name)
        fileB = open(str(config_file_name), 'w')
        fileB.write('Last modified: ' + time + '\n')
        fileB.write('Program ' + self.version)
        fileB.write(' Program counter: ' + str(self.progr_count - 1) + '\n')
        fileB.write('Tab 1' + '\n')
        fileB.write(' Initial directory: ' + str(self.initialdirectory) + '\n')
        fileB.write(' Pxl ratio: ' + str(self.pxl_ratio) + '\n')
        fileB.write(' Used Threshold: ' + str(self.thresholding_var.get()) + '\n')
        fileB.write(' Which Threshold: ' + str(self.type_threshold) + '\n')
        fileB.write('Tab 2' + '\n')
        fileB.write(' Border? ' + str(self.particles_at_border) + '\n')
        fileB.write(' Sigma: ' + str(self.sigma) + '\n')
        fileB.write(' Low Threshold: ' + str(self.low_threshold) + '\n')
        fileB.write(' High Threshold: ' + str(self.high_threshold) + '\n')
        fileB.write(' LabelsLine Parameter: ' + str(self.labelLines_parameter) + '\n')
        fileB.write('Areas were detected and refined at ')
        fileB.write('Tab 3' + '\n')
        fileB.write(' After filtering with \n')
        fileB.write(str(self.filter_used_list))
        fileB.write(', those ellipsoids were drawn on the image \n')
        for i in self.ellipse_draw_list:
            fileB.write('%s\n' % i)
        fileB.write('maybe some were then kicked by hand to get the list in the file: evaluation')
        fileB.close()

        self.progressbar.step(24)

        labelo = tk.Label(master=self.save_sucess, text='Saving successful!')
        labelo.pack(pady=10, padx=10)
        butt_save_win = ttk.Button(master=self.save_sucess, text='OK', command=lambda: self.save_sucess.destroy())
        butt_save_win.pack(pady=10, padx=10)

    # load the config-data of previous evaluated images
    def load_evaluation_parameter(self):
        self.ev_filename = filedialog.askopenfilename(initialdir=self.initialdirectory, title='Select evaluation file', filetypes=(('Evaluation config', '*.txt'), ('All files', '*.*')))
        ev = str(self.ev_filename)
        # catch error if filedialog is closed without choosing a file
        if ev:
            file_ev = open(ev, 'r')
            ev_content = file_ev.readlines()
            file_ev.close()
            if self.console_print_on: print(ev_content)
            k = 0
            for i in ev_content:
                if 'Pxl ratio:' in i:
                    if self.console_print_on: print(ev_content[4])
                    self.pxl_ratio = float(i.split(': ')[1])
                    if self.console_print_on: print(self.pxl_ratio)
                    k = 1
                if ' Border? ' in i:
                    particles_at_border = str(i.split('? ')[1])
                    print(particles_at_border)
                    k = 1
                if ' Sigma: ' in i:
                    self.sigma = float(i.split(': ')[1])
                    k = 1
                if ' Low Threshold: ' in i:
                    self.low_threshold = float(i.split(': ')[1])
                    k = 1
                if ' High Threshold: ' in i:
                    self.high_threshold = float(i.split(': ')[1])
                    k = 1
                if ' LabelsLine Parameter: ' in i:
                    self.labelLines_parameter = float(i.split(': ')[1])
                    k = 1
            if k == 1:
                if self.console_print_on: print('some parameters from file successfully loaded')
            else:
                if self.console_print_on: print('no parameter in file')
        if self.progr_count == 2:
            # update pxl ratio
            self.content_tab1.update_pxlratio_tab1(self.pxl_ratio)
        else:
            print(self.progr_count)

    # save window that replies successful data saved with status bar
    def save_window(self):
        self.save_sucess = tk.Tk()
        self.save_sucess.geometry('140x120')
        self.progressbar = ttk.Progressbar(master=self.save_sucess, length=200, mode='determinate', maximum=100)
        self.progressbar.pack()
        self.save_sucess.focus_force()
        self.save_sucess.lift()

    # credits window
    def credits_win(self):
        credits = tk.Tk()
        credits.geometry('140x70')
        label1 = tk.Label(master=credits, text=self.version)
        label1.pack(padx=10)
        label2 = tk.Label(master=credits, text='Copyright 2024')
        label2.pack(pady=5, padx=10)
        label3 = tk.Label(master=credits, text='Dominik Benke, M.Sc.')
        label3.pack(padx=10)

    # load image to the program
    def load_image(self):
        if self.console_print_on: print(self.automated_crop.get())
        # open image file and parse it to self.original and self.image_rgb
        self.filename = filedialog.askopenfilename(initialdir=self.initialdirectory, title='Select file', filetypes=(('Images', '*.tif *.bmp *.jpg'), ('All files', '*.*'), ('Tiffs', '*.tif'), ('Bitmaps', '*.bmp'), ('Jpegs', '*.jpg')))
        a = str(self.filename)
        # catch error if filedialog is closed without choosing a file
        # dummypic is used
        if a == '':
            a = self.dummypic
        # save image path at global variables
        self.image_path = os.path.split(os.path.abspath(a))
        self.initialdirectory = self.image_path[0]
        if self.console_print_on: print('The image directory: ' + str(self.image_path))
        self.file_name = self.image_path[1]
        if self.console_print_on: print('The image name: ' + str(self.image_path[1]))
        # add imagename to the menu bar
        self.menubar.add_command(label=self.file_name, command=lambda: self.open_new(self.tabControl))
        self.config(menu=self.menubar)

        ## image read and crop ##
        # #global original
        self.original = io.imread(a)
        # #global imagesize
        self.imagesize = self.original.shape
        if self.console_print_on: print(self.imagesize[0])
        if self.console_print_on: print(self.imagesize[1])
        # if automated_crop =1 (ON, toggle at menu): automated cutting of the annotation in the img 
        # by finding the first horizontal pixel row with highest color black 
        # (value 0, but not all pixel have this) and cutting picture there
        self.cutoff_pic = 0
        if self.automated_crop.get():
            self.grey_complete = img_as_ubyte(io.imread(a, as_gray=True))
            if self.console_print_on: print(self.original[689:693])
            if self.console_print_on: print(self.grey_complete[689:693])
            blackline_y_list = []
            for y in self.grey_complete:
                line_add = 0
                for x in y:
                    if x < 10:
                        line_add = line_add + 1
                blackline_y_list.append(line_add)
            if self.console_print_on: print(blackline_y_list)
            if self.console_print_on: print('The maximum of black pxl in a line is at position: ' + str(np.argmax(blackline_y_list)))
            self.cutoff_pic = int(np.argmax(blackline_y_list)-1)
        # if no_crop is ON (toggle at menu) use full size of image
        # if no_crop is OFF the automated crop or a standard crop at 90% will be done.
        # 90% crop = crop standard space of annotations at University Bayreuths SEM images from ZEISS microscopes
        if self.no_crop.get():
            y = self.imagesize[0]
        else:
            if self.cutoff_pic<10:
                y = int(round(self.imagesize[0]*0.9, 0))
            else:
                y = self.cutoff_pic
        self.image_rgb = io.imread(a, as_gray = True)[:y]
        if self.console_print_on: print(self.image_rgb.shape)
        if self.console_print_on: print(self.image_rgb)
        # transform rgb image to appropriate color/grey value format
        self.image_rgb = img_as_ubyte(self.image_rgb)
        self.image_loaded = self.image_rgb.copy()
        if self.console_print_on: print(self.image_rgb.shape)
        if self.console_print_on: print(self.image_rgb)        
        # global new_imagesize array of (height, width)
        self.new_imagesize = self.image_rgb.shape
        if self.console_print_on: print(self.image_rgb.shape)

    # new dummys for (re-)start evaluation
    def new_dummys(self):
        if self.console_print_on: print('new dummys')
        ## dummys for global variables ##
        self.image_rgb = color.rgb2gray(io.imread(self.dummypic)[:630])
        self.original = []
        # imagesize and new_imagesize are arrays of (height, width)
        self.imagesize = self.image_rgb.shape
        self.new_imagesize = self.image_rgb.shape
        self.edges = feature.canny(self.image_rgb)
        self.gradient = rank.gradient(img_as_ubyte(self.edges), disk(3)) < 10
        self.markers = ndi.label(self.gradient)[0]
        self.labels = watershed(self.edges, self.markers)
        self.ParticleProp = regionprops(self.labels)
        self.particles_at_border = False
        self.image_path = []
        # np.zeros need (height, width), so using new_imagesize in normal order is OK
        self.image_ellipses = np.zeros([self.new_imagesize[0], self.new_imagesize[1], 3])
        self.image_numbers = self.image_ellipses
        self.ellipse_draw_list = []
        self.ParticleList = []
        self.final_ParticleList = []
        self.WorkingList = []


# start main application
def main():
    app = EllipsoidDetectionApp()
    app.mainloop()


if __name__ == '__main__':
    main()
