################################
# ############################ #
# ### Tab 1 - Start Page - ### #
# ############################ #
################################

# import packages
import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.figure import Figure
from matplotlib.pyplot import rcParams
import tkinter as tk
from tkinter import ttk
from skimage.filters import threshold_otsu, threshold_local

# global plot parameters
matplotlib.use("TkAgg")
LARGE_FONT = ("Verdana", 12)
GlobFontSize = 10
params = {
    'axes.labelsize': GlobFontSize,
    'font.size': GlobFontSize,
    'axes.titlesize': GlobFontSize,
    'font.family': "Calibri",
    'legend.fontsize': GlobFontSize,
    'xtick.labelsize': GlobFontSize,
    'ytick.labelsize': GlobFontSize,
    'text.usetex': False,
    'figure.constrained_layout.use': True,
    'lines.markersize': 3.0,
    'boxplot.flierprops.markersize': 3
}
rcParams.update(params)

# tab 1 code
class StartPage(tk.Frame):
    def __init__(self, parent, controller):
        # get variables from controller
        self.console_print_on = controller.console_print_on
        self.image_rgb = controller.image_rgb
        self.original = controller.original
        self.img_x = controller.img_x
        self.img_y = controller.img_y
        self.thresholding_var = controller.thresholding_var.get()
        self.image_w_axes = controller.image_w_axes.get()

        # make working copy of greyscale & resized image
        self.image = self.image_rgb

        ## create tab layout ##
        tk.Frame.__init__(self, parent)
        # # devide window into 2 frames
        tab1_left = ttk.Frame(master=self)
        tab1_left.grid(row=0, column=0, sticky="nw")
        tab1_left.columnconfigure(0, minsize=220)
        self.tab1_right = ttk.Frame(master=self, relief="groove")
        self.tab1_right.grid(row=0, column=1, sticky="nsew")

        ## make menu on the left of tab 1 with buttons and field for pxl ratio ##
        butt1_2 = ttk.Button(master=tab1_left, text="Use Threshold", command=lambda: self.thresholding_images(controller))
        butt1_2.grid(row=0, column=0, pady=10, sticky='ew')
        label = tk.Label(master=tab1_left, text="Pixel ratio in nm/pxl from SEM image \n (use dot, not comma):")
        label.grid(row=1, column=0, rowspan=1)
        self.entryfield_pxlratio = tk.Entry(master=tab1_left)
        self.entryfield_pxlratio.grid(row=2, column=0, rowspan=1)
        self.entryfield_pxlratio.insert(0, str(controller.pxl_ratio))
        ttk.Separator(master=tab1_left, orient='horizontal').grid(row=3, column=0, pady=10, sticky="ew")
        butt1 = ttk.Button(master=tab1_left, text=">> Go to Canny edge detection", command=lambda: self.update_controller_from_tab1(controller))
        butt1.grid(row=4, column=0, sticky='ew')

        # display images on right side of tab 1
        self.show_original_image()
        self.show_gray_image()
        # if wanted one can threshold the image and choose the appropriate image to go further
        if self.thresholding_var:
            self.thresholding_images(controller)

    # show original image on tab 1
    def show_original_image(self):
        fig1 = Figure(figsize=(self.img_x, self.img_y + 2))
        a = fig1.add_subplot(111)
        a.imshow(self.original)
        if self.image_w_axes:
            a.set_title("Original")
            a.set_xlabel("Horizontal Pixel ID")
            a.set_ylabel("Vertical Pixel ID")
        canvas1_1 = FigureCanvasTkAgg(fig1, master=self.tab1_right)
        canvas1_1.get_tk_widget().grid(row=0, column=0, rowspan=1)
        # toolbar
        toolbar = NavigationToolbar2Tk(canvas1_1, self.tab1_right, pack_toolbar=False)
        toolbar.update()
        toolbar.grid(row=1, column=0, rowspan=1)

    # show converted to grayscale and resized image
    def show_gray_image(self):
        fig2 = Figure(figsize=(self.img_x, self.img_y + 2))
        a2 = fig2.add_subplot(111)
        a2.imshow(self.image, cmap="gray")
        a2.set_title("Grayscale & Resized")
        a2.set_xlabel("Horizontal Pixel ID")
        a2.set_ylabel("Vertical Pixel ID")
        canvas1_2 = FigureCanvasTkAgg(fig2, master=self.tab1_right)
        canvas1_2.get_tk_widget().grid(row=0, column=1, rowspan=1, columnspan=2)
        # toolbar
        toolbar = NavigationToolbar2Tk(canvas1_2, self.tab1_right, pack_toolbar=False)
        toolbar.update()
        toolbar.grid(row=1, column=1, rowspan=1)

    # thresholding the image and show result in a new window
    def thresholding_images(self, controller):
        self.threshold_window = tk.Tk()

        # local thresholding
        block_size = 35
        local_thresh = threshold_local(self.image, block_size)
        self.binary_local = self.image < local_thresh
        fig = Figure(figsize=(self.img_x, self.img_y))
        thres_plot = fig.add_subplot(111)
        thres_plot.imshow(self.binary_local, cmap="gray")
        thres_plot.set_title("local thresholding")
        thres_plot.set_xlabel("Horizontal Pixel ID")
        thres_plot.set_ylabel("Vertical Pixel ID")
        canvas3 = FigureCanvasTkAgg(fig, self.threshold_window)
        canvas3.get_tk_widget().grid(row=3, column=0, rowspan=1)

        # global thresholding (otsu)
        global_thresh = threshold_otsu(self.image)
        self.binary_global = self.image < global_thresh
        figa = Figure(figsize=(self.img_x, self.img_y))
        thresplot2 = figa.add_subplot(111)
        thresplot2.imshow(self.binary_global, cmap="gray")
        thresplot2.set_title("global thresholding")
        thresplot2.set_xlabel("Horizontal Pixel ID")
        thresplot2.set_ylabel("Vertical Pixel ID")
        canvas4 = FigureCanvasTkAgg(figa, self.threshold_window)
        canvas4.get_tk_widget().grid(row=3, column=1, rowspan=1, columnspan=2)

        # chose which threshold to use by buttons
        ttk.Button(self.threshold_window, text="Use binary local", command=lambda: self.checkbx_local(controller)).grid(row=4, column=0)
        ttk.Button(self.threshold_window, text="Use binary global", command=lambda: self.checkbx_global(controller)).grid(row=4, column=1)

    # routine after click on button for local thresholding
    def checkbx_local(self, controller):
        self.image = self.binary_local
        # new initial parameters for sliders at tab 2
        controller.sigma = 2
        controller.low_threshold = 0
        controller.high_threshold = 1
        controller.labelLines_parameter = 2
        controller.thresholding_var.set(True)
        controller.type_threshold = 'binary_local'
        self.threshold_window.destroy()

    # routine after click on button for global thresholding
    def checkbx_global(self, controller):
        self.image = self.binary_global
        # new initial parameters for sliders at tab 2
        controller.sigma = 3.0
        controller.low_threshold = 0
        controller.high_threshold = 1
        controller.labelLines_parameter = 2
        controller.thresholding_var.set(True)
        controller.type_threshold = 'binary_global_otsu'
        self.threshold_window.destroy()

    # routine after click on button '>> Go to Canny edge detection'
    def update_controller_from_tab1(self, controller):
        # hadover of working copy of image to controller
        controller.image = self.image
        # read pxl_ratio
        f = self.entryfield_pxlratio.get()
        try:
            controller.pxl_ratio = float(f)
            controller.next_step(2)
            controller.show_frame(1)
        # if wrong/no number in entry field
        except ValueError:
            if self.console_print_on: print("oops, default: " + str(controller.pxl_ratio))
            self.error_catcher(controller)
        if self.console_print_on: print(f)

    # handle exception of wrong number in entry field
    def error_catcher(self, controller):
        self.error_frame = tk.Tk()
        self.error_frame.geometry("220x100")
        label_text = "wrong value, please write new"
        label = tk.Label(master=self.error_frame, text=label_text)
        label.pack(pady=10, padx=10)
        self.newentry = tk.Entry(master=self.error_frame)
        self.newentry.pack()
        butt1 = tk.Button(master=self.error_frame, text="read", command=lambda: self.try_read_again(controller))
        butt1.pack()

    # handle exception if still a wrong number in entry field
    def try_read_again(self, controller):
        try:
            controller.pxl_ratio = float(self.newentry.get())
            if self.console_print_on: print("New value for pxl is " + str(controller.pxl_ratio))
            self.error_frame.destroy()
            controller.next_step(2)
            controller.show_frame(1)
        except ValueError:
            tk.Label(master=self.error_frame, text="try again with dot, no comma").pack()

    # routine executed when saving
    def make_save_update_from_tab1(self, controller):
        controller.image = self.image
        # read pxl_ratio again
        f = self.entryfield_pxlratio.get()
        try:
            controller.pxl_ratio = float(f)
        except ValueError:
            if self.console_print_on: print("oops, default: " + str(controller.pxl_ratio))
        if self.console_print_on: print(f)

    # update pxl ratio when config-data is loaded for re-evaluation
    def update_pxlratio_tab1(self, pxl_new):
        self.entryfield_pxlratio.delete(0, 'end')
        self.entryfield_pxlratio.insert(0, str(pxl_new))
        if self.console_print_on: print('A new pxl ratio is loaded: ' + str(pxl_new))
