##########################################
# ###################################### #
# ### Tab 2 - Canny Edge Detection - ### #
# ###################################### #
##########################################

# import packages
import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.figure import Figure
from matplotlib.pyplot import rcParams
import matplotlib.pyplot as plt
import numpy as np
import tkinter as tk
from tkinter import ttk
from skimage import feature, img_as_ubyte
from skimage.filters import rank
from skimage.segmentation import watershed
from skimage.morphology import disk
from skimage.measure import regionprops
from scipy import ndimage as ndi

# global plot parameters
matplotlib.use("TkAgg")
LARGE_FONT = ("Verdana", 12)
GlobFontSize = 10
params = {
    'axes.labelsize': GlobFontSize,
    'font.size': GlobFontSize,
    'axes.titlesize': GlobFontSize,
    'font.family': "Calibri",
    'legend.fontsize': GlobFontSize,
    'xtick.labelsize': GlobFontSize,
    'ytick.labelsize': GlobFontSize,
    'text.usetex': False,
    'figure.constrained_layout.use': True,
    'lines.markersize': 3.0,
    'boxplot.flierprops.markersize': 3
}
rcParams.update(params)


# tab 2 code
class Page_Area_detection(tk.Frame):
    def __init__(self, parent, controller):
        # get variables from controller
        self.console_print_on = controller.console_print_on
        self.particles_at_border = controller.particles_at_border
        self.image = controller.image
        self.image_loaded = controller.image_rgb
        self.img_x = controller.img_x
        self.img_y = controller.img_y
        self.sigm = controller.sigma
        self.high_threshold = controller.high_threshold
        self.low_threshold = controller.low_threshold
        self.labelLines_parameter = controller.labelLines_parameter
        self.image_w_axes = controller.image_w_axes.get()
        # imagesize is (height, width)
        self.new_imagesize = controller.new_imagesize

        ## create tab layout ##
        tk.Frame.__init__(self, parent)
        # devide window into 2 frames
        tab2_left = ttk.Frame(master=self)
        tab2_left.grid(row=0, column=0, sticky="nw")
        tab2_left.columnconfigure(0, minsize=220)
        if self.console_print_on: print("show particles at border: ")
        if self.console_print_on: print(controller.particles_at_border)        
        self.tab2_right = ttk.Frame(master=self)
        self.tab2_right.grid(row=0, column=1, sticky="nsew")
        self.tab2_right.update()
        if self.console_print_on: print("size of tab2_right")
        if self.console_print_on: print(self.tab2_right.winfo_width())

        # make tab 2 menu on left container
        # # Different variables for tab menu
        sc_sigm = tk.DoubleVar()
        sc_b = tk.DoubleVar()
        sc_c = tk.DoubleVar()
        sc_param = tk.DoubleVar()
        self.checkbx1_var = tk.BooleanVar()
        self.checkbx2_var = tk.BooleanVar()
        self.checkbx2_var.set(False)

        # # create checkbuttons
        ttk.Checkbutton(tab2_left, text="Simultaneous Plotting", variable=self.checkbx1_var, command=self.checkbx1).grid(row=0, column=0, sticky="ew")
        ttk.Checkbutton(tab2_left, text="Add Particles at Borders", variable=self.checkbx2_var, command=self.checkbx2).grid(row=1, column=0, sticky="ew")
        if self.console_print_on: print("Particles at borders used? " + str(self.checkbx2_var.get()))
        # # variable sigma
        ttk.Label(tab2_left, text="Sigma").grid(row=2, column=0, sticky="ew")
        self.scale_sigma = tk.Scale(tab2_left, variable=sc_sigm, from_=0.00, to=10.00, length=200, digits=3, orient=tk.HORIZONTAL, resolution=0.025)
        if self.console_print_on: print("width scale")
        if self.console_print_on: print(self.scale_sigma.cget("width"))
        self.scale_sigma.set(self.sigm)
        self.scale_sigma.grid(row=3, column=0, sticky="ew")
        # # show the grey value histogram on left side
        self.fig_h = Figure(figsize=(2, 2))
        hist_plot = self.fig_h.add_subplot(111)
        hist_plot.clear()
        hist_plot.hist(self.image.ravel(), 256, [0, 256])
        canvas2_h1 = FigureCanvasTkAgg(self.fig_h, tab2_left)
        canvas2_h1.get_tk_widget().grid(row=4, column=0, rowspan=1)
        # # create sliders for threshold
        ttk.Label(tab2_left, text="low_threshold").grid(row=5, column=0, sticky="ew")
        self.scale_b = tk.Scale(tab2_left, variable=sc_b, from_=0.00, to=256.0, length=200, digits=2, orient=tk.HORIZONTAL, resolution=0.05)
        self.scale_b.set(self.low_threshold)
        self.scale_b.grid(row=6, column=0, sticky="ew")
        ttk.Label(tab2_left, text="high_threshold").grid(row=7, column=0, sticky="ew")
        self.scale_c = tk.Scale(tab2_left, variable=sc_c, from_=0.00, to=256.0, length=200, digits=2, orient=tk.HORIZONTAL, resolution=0.05)
        self.scale_c.set(self.high_threshold)
        self.scale_c.grid(row=8, column=0, sticky="ew")
        # # create vertical lines on the histogram to adjust for edge detection
        self.lower_limit_line = hist_plot.axvline(self.scale_b.get(), color='k')
        self.upper_limit_line = hist_plot.axvline(self.scale_c.get(), color='k')
        self.scale_b.config(command=self.update_histo)
        self.scale_c.config(command=self.update_histo)
        # # button for edge detection
        butt2_1 = ttk.Button(tab2_left, text="> Edge Detection", command=self.scroll_canny_detection)
        butt2_1.grid(row=9, column=0, sticky="ew")
        # # create slider for line parameter
        ttk.Label(tab2_left, text="Labels - Line Parameter").grid(row=11, column=0, sticky="ew", pady=(10, 0))
        self.scale_param = tk.Scale(tab2_left, variable=sc_param, from_=0.00, to=10.00, digits=1, orient=tk.HORIZONTAL, resolution=1)
        self.scale_param.set(self.labelLines_parameter)
        self.scale_param.config(state="disabled")
        self.scale_param.grid(row=12, column=0, sticky="ew")
        # # button for label image
        self.butt2_3 = ttk.Button(tab2_left, text="> Labels", command=self.plot_markers, state="disabled")
        self.butt2_3.grid(row=13, column=0, sticky="ew")
        # # button for watershed
        self.butt2_4 = ttk.Button(tab2_left, text="> Watershed Separation", command=self.watershedsep, state="disabled")
        self.butt2_4.grid(row=14, column=0, sticky="ew", pady=(10, 0))
        ttk.Separator(tab2_left, orient='horizontal').grid(row=15,column=0, pady=10, sticky="ew")
        # # button for perform all steps at once
        butt2_6 = ttk.Button(tab2_left, text="DO ALL", command=self.all_do)
        butt2_6.grid(row=16, column=0, sticky="ew")
        ttk.Separator(tab2_left, orient='horizontal').grid(row=17,column=0, pady=10, sticky="ew")
        # # button for next tab
        self.butt2_5 = ttk.Button(tab2_left, text=">> Go to Evaluation", command=lambda: self.update_controller_from_tab2(controller), state="disabled")
        self.butt2_5.grid(row=18, column=0, sticky="ew")

        tab2_left.update()
        if self.console_print_on: print("size of tab2_left")
        if self.console_print_on: print(tab2_left.winfo_width())
        # variables for updating tab during replotting images
        self.canny_drawn = 0
        self.marker_drawn = 0
        self.watershed_drawn = 0

    # Update the position of the vertical lines in the histogram on left
    def update_histo(self, value):
        self.lower_limit_line.set_xdata([self.scale_b.get(), self.scale_b.get()])
        self.upper_limit_line.set_xdata([self.scale_c.get(), self.scale_c.get()])
        if self.console_print_on: print("updated histo")
        self.fig_h.canvas.draw_idle()

    # perform all steps at once
    def all_do(self):
        self.scroll_canny_detection()
        self.plot_markers()
        self.watershedsep()

    # in checkbox 1 you can say if you want interactive plotting
    def checkbx1(self):
        if self.checkbx1_var.get() is True:
            if self.console_print_on: print("Interactive mode: ON")
            self.scale_sigma.config(command=self.simultane_canny)
            self.scale_b.config(command=self.simultane_canny)
            self.scale_c.config(command=self.simultane_canny)
            self.scale_param.config(command=self.simultane_plot_markers)
        else:
            if self.console_print_on: print("Interactive mode: OFF")
            self.scale_sigma.config(command=self.no_action)
            self.scale_b.config(command=self.update_histo)
            self.scale_c.config(command=self.update_histo)
            self.scale_param.config(command=self.no_action)

    # if checkbox 1 is on, canny image is plotted when slider value is changed
    def simultane_canny(self, a):
        self.update_histo(a)
        self.plot_Canny(a)
        if not self.marker_drawn == 0:
            self.plot_markers()
            if not self.watershed_drawn == 0:
                self.watershedsep()

    # if checkbox 1 is on, marker image is plotted when slider value is changed
    def simultane_plot_markers(self, value):
        self.plot_markers()
        if not self.watershed_drawn == 0:
            self.watershedsep()

    # checkbox 2 let you choose if you want to take bordered particles into account
    def checkbx2(self):
        if self.checkbx2_var.get() is True:
            self.particles_at_border = True
            if self.console_print_on: print("Particles at border: ON")
        else:
            self.particles_at_border = False
            if self.console_print_on: print("Particles at border: OFF")

    # method needed as the sliders command cannot be turned off elseway
    def no_action(self, value):
        unused_var = 1
        if self.console_print_on: print(value)

    # method needed as the sliders will hand over the slider value; and canny does not need one
    def plot_Canny(self, value):
        self.scroll_canny_detection()
        if self.console_print_on: print("Current slider value:")
        if self.console_print_on: print(value)

    # detect borders of particles/areas by using canny of skimage package
    def scroll_canny_detection(self):
        # get slider variables
        self.sigm = float(str(self.scale_sigma.get()))
        self.low_threshold = float(str(self.scale_b.get()))
        self.high_threshold = float(str(self.scale_c.get()))
        # make canny
        self.edges = feature.canny(self.image, sigma=self.sigm, low_threshold=self.low_threshold, high_threshold=self.high_threshold)
        # create images with the detected particle/area borders
        # check if images are already drawn, create if not, update if yes
        if self.canny_drawn == 0:
            # show the canny image
            self.fig_canny = Figure(figsize=(self.img_x, self.img_y))
            canny_plot = self.fig_canny.add_subplot(111)
            canny_plot.clear()
            self.ca_plot = canny_plot.imshow(self.edges, cmap="gray", interpolation="nearest")
            if self.image_w_axes:
                canny_plot.axis("on")
                canny_plot.set_title("Canny")
                canny_plot.set_xlabel("Horizontal Pixel ID")
                canny_plot.set_ylabel("Vertical Pixel ID")
            else:
                canny_plot.axis("off")
            canvas2_1 = FigureCanvasTkAgg(self.fig_canny, self.tab2_right)
            canvas2_1.get_tk_widget().grid(row=0, column=0, rowspan=1)
            # toolbar
            toolbar = NavigationToolbar2Tk(canvas2_1, self.tab2_right, pack_toolbar=False)
            toolbar.update()
            toolbar.grid(row=1, column=0, rowspan=1)

            # show canny on grey scale image
            self.fig2_2 = Figure(figsize=(self.img_x, self.img_y))
            d = self.fig2_2.add_subplot(111)
            self.ca_on_img = d.imshow(self.edges, cmap=plt.cm.hsv)
            d.imshow(self.image, cmap=plt.cm.gray, alpha=0.5)
            if self.image_w_axes:
                d.axis("on")
                d.set_title("Canny on Original")
                d.set_xlabel("Horizontal Pixel ID")
                d.set_ylabel("Vertical Pixel ID")
            else:
                d.axis("off")
            canvas2_2 = FigureCanvasTkAgg(self.fig2_2, self.tab2_right)
            canvas2_2.get_tk_widget().grid(row=0, column=1, rowspan=1)
            # toolbar
            toolbar2 = NavigationToolbar2Tk(canvas2_2, self.tab2_right, pack_toolbar=False)
            toolbar2.update()
            toolbar2.grid(row=1, column=1, rowspan=1)

            self.canny_drawn = 1
            self.fig_canny.canvas.draw_idle()
            self.fig2_2.canvas.draw_idle()

        else:
            # update canny and on grey scale image when self.edges altered
            self.ca_plot.set_data(self.edges)
            self.ca_on_img.set_data(self.edges)
            self.fig_canny.canvas.draw_idle()
            self.fig2_2.canvas.draw_idle()
            if self.console_print_on: print("canny images updated")

        self.tab2_right.update()
        if self.console_print_on: print("size of tab2_right")
        if self.console_print_on: print(self.tab2_right.winfo_width())
        if self.console_print_on: print("Canny Image Plotted") 
        # enable next step
        self.butt2_3.config(state="normal")
        self.scale_param.config(state="normal")

    # create marker image from slider value for different areas
    def plot_markers(self):
        # global gradient via rank
        self.labelLines_parameter = float(str(self.scale_param.get()))
        self.gradient = rank.gradient(img_as_ubyte(self.edges), disk(self.labelLines_parameter)) < 10
        # global markers
        self.markers = ndi.label(self.gradient)[0]
        if self.console_print_on: print(ndi.label(self.gradient)[0])
        newcmp = 'YlOrRd'
        # check if image is already drawn, create if not, update if yes
        if self.marker_drawn == 0:
            # show the marker image 1st time
            self.fig2_3 = Figure(figsize=(self.img_x, self.img_y))
            marker_img = self.fig2_3.add_subplot(111)
            self.marker_img_update = marker_img.imshow(self.markers, cmap=newcmp)
            if self.image_w_axes:
                marker_img.set_title("Markers")
                marker_img.set_xlabel("Horizontal Pixel ID")
                marker_img.set_ylabel("Vertical Pixel ID")
            canvas2_3 = FigureCanvasTkAgg(self.fig2_3, self.tab2_right)
            canvas2_3.get_tk_widget().grid(row=2, column=0, rowspan=1)
            # toolbar
            toolbar2_3 = NavigationToolbar2Tk(canvas2_3, self.tab2_right, pack_toolbar=False)
            toolbar2_3.update()
            toolbar2_3.grid(row=3, column=0, rowspan=1)
            if self.console_print_on: print("Marker Image Plotted")
            self.marker_drawn = 1

        else:
            # update marker image and displayed label color when scale_param altered
            self.marker_img_update.set_data(self.markers)
            self.marker_img_update.set_clim(vmax=np.unique(self.markers).size)
            self.fig2_3.canvas.draw_idle()
            if self.console_print_on: print("marker image updated")
        # enable next step
        self.butt2_4.config(state="normal")

    # perform watershed separation
    def watershedsep(self):
        self.labels = watershed(self.edges, self.markers)
        # check if image is already drawn, create if not, update if yes
        if self.watershed_drawn == 0:
            # show the watershed image
            self.fig2_4 = Figure(figsize=(self.img_x, self.img_y))
            a = self.fig2_4.add_subplot(111)
            self.watershed_img_update = a.imshow(self.labels, cmap='prism')
            if self.image_w_axes:
                a.set_title("Watershed Labels")
                a.set_xlabel("Horizontal Pixel ID")
                a.set_ylabel("Vertical Pixel ID")
            canvas2_4 = FigureCanvasTkAgg(self.fig2_4, self.tab2_right)
            canvas2_4.get_tk_widget().grid(row=2, column=1, rowspan=1)
            # toolbar
            toolbar2_4 = NavigationToolbar2Tk(canvas2_4, self.tab2_right, pack_toolbar=False)
            toolbar2_4.update()
            toolbar2_4.grid(row=3, column=1, rowspan=1)
            if self.console_print_on: print("Watershed Separation Plotted")
            self.watershed_drawn = 1
        else:
            self.watershed_img_update.set_data(self.labels)
            self.watershed_img_update.set_clim(vmax=np.unique(self.labels).size)
            self.fig2_4.canvas.draw_idle()
            if self.console_print_on: print("watershed image updated")
        # enable next step
        self.butt2_5.config(state="normal")

    # read out of properties of the detected areas
    def create_ParticleProp_list(self):
        self.ParticleProp = regionprops(self.labels)
        # new ParticleProp_List if no border particles should be taken into account,
        # else unaltered ParticleProp_List is used
        if self.particles_at_border is False:
            if self.console_print_on: print("Do remove areas/particles at borders")
            new_ParticleProp = []
            for P in self.ParticleProp:
                # self.new_imagesize[0] == height ; self.new_imagesize[1] == width
                # Bounding box (min_row, min_col, max_row = height, max_col = width)
                if P.bbox[0] > 2 and P.bbox[1] > 2 and P.bbox[2] < (self.new_imagesize[0] - 2) and P.bbox[3] < (self.new_imagesize[1] - 2):
                    new_ParticleProp.append(P)
            self.ParticleProp = new_ParticleProp

        # clean particle prop from minor and major axis == 0
        new_list = []
        zero_list = []
        for P in self.ParticleProp:
            if not P.major_axis_length == 0 and not P.minor_axis_length == 0:
                new_list.append(P)
            else:
                zero_list.append(P)
        self.ParticleProp = new_list

        if self.console_print_on:
            for item in zero_list:
                print('Particle with zero dimension: No. {0} with major {1} and minor {2}'.format(item.label, item.major_axis_length, item.minor_axis_length))

    # update images and then handover the altered parameters to the controller, so they can be used with next tab
    def update_controller_from_tab2(self, controller):
        self.all_do()
        self.create_ParticleProp_list()
        # variables/pictures/list which are made/altered by window
        controller.edges = self.edges        # edges of areas
        controller.gradient = self.gradient  # gradient used
        controller.markers = self.markers    # markers of areas
        controller.labels = self.labels      # image of separeted areas
        controller.ParticleProp = self.ParticleProp  # region properties of separated areas
        # checkbox variable
        controller.particles_at_border = self.particles_at_border
        # current values of sliders
        controller.sigma = self.sigm
        controller.high_threshold = self.high_threshold
        controller.low_threshold = self.low_threshold
        controller.labelLines_parameter = self.labelLines_parameter
        # go to next tab 3
        controller.next_step(3)
        controller.show_frame(2)

    # handover when saving the parameters
    def make_save_update_from_tab2(self, controller):
        # checkbox variable
        controller.particles_at_border = self.particles_at_border
        # current values of sliders
        controller.sigma = self.sigm
        controller.high_threshold = self.high_threshold
        controller.low_threshold = self.low_threshold
        controller.labelLines_parameter = self.labelLines_parameter
