#############################################
# ######################################### #
# ### Tab 3 - Ellipse areas selection - ### #
# ######################################### #
#############################################

# import packages
import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.figure import Figure
from matplotlib.pyplot import rcParams
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import tkinter as tk
from tkinter import ttk
from skimage.draw import ellipse
import math
from PIL import Image, ImageDraw, ImageFont

# global plot parameters
matplotlib.use("TkAgg")
LARGE_FONT = ("Verdana", 12)
GlobFontSize = 10
params = {
    'axes.labelsize': GlobFontSize,
    'font.size': GlobFontSize,
    'axes.titlesize': GlobFontSize,
    'font.family': "Calibri",
    'legend.fontsize': GlobFontSize,
    'xtick.labelsize': GlobFontSize,
    'ytick.labelsize': GlobFontSize,
    'text.usetex': False,
    'figure.constrained_layout.use': True,
    'lines.markersize': 3.0,
    'boxplot.flierprops.markersize': 3
}
rcParams.update(params)


# tab 3 code
class Page_Area_Selection(tk.Frame):
    def __init__(self, parent, controller):
        # get dummys and variables from controller
        self.console_print_on = controller.console_print_on
        self.pxl_ratio = controller.pxl_ratio
        self.ParticleProp = controller.ParticleProp
        self.new_imagesize = controller.new_imagesize  # imagesize is (height, width)
        self.particles_at_border = controller.particles_at_border
        self.ParticleList = controller.ParticleList
        self.ellipse_draw_list = controller.ellipse_draw_list  # narrowed down particles img pxl list
        self.image_ellipses = controller.image_ellipses
        self.image_numbers = controller.image_numbers
        self.image = controller.image
        self.image_w_axes = controller.image_w_axes.get()
        self.img_x = controller.img_x
        self.img_y = controller.img_y

        # start variables of the tab for the entryBoxes
        self.highest_area = 1000000
        self.minimum_area = 0
        self.highest_major = 10000
        self.minimum_major = 0
        self.highest_minor = 10000
        self.minimum_minor = 0
        self.highest_aspect = 100
        self.minimum_aspect = 0
        self.highest_orientation = 180
        self.minimum_orientation = 0

        # other variables
        self.ellipse_color = plt.cm.hsv
        self.area_factor = self.pxl_ratio * self.pxl_ratio / 1  # in nm^2
        self.reset_ParticleProp_list = None
        self.ellpsoids_at_tab3_drawn = 0
        self.max_aspect = 0.0  # for polarplot

        ## create tab layout ##
        tk.Frame.__init__(self, parent)
        # devide window into 2 frames
        self.tab3_left = tk.Frame(master=self)
        self.tab3_left.grid(row=1, column=0, sticky="nw")
        self.tab3_left.columnconfigure(0, minsize=220)
        self.tab3_right = tk.Frame(master=self)
        self.tab3_right.grid(row=1, column=1, sticky="ne")

        # make tab 3 menu on left container
        # checkbutton on left, which value to select
        self.rad_but_var = tk.IntVar()
        self.R1 = tk.Radiobutton(self.tab3_left, text="Area", variable=self.rad_but_var, value=1, command=self.radiobut_chosen)
        self.R1.grid(row=0, column=0, sticky="nw")
        self.R2 = tk.Radiobutton(self.tab3_left, text="Major", variable=self.rad_but_var, value=2, command=self.radiobut_chosen)
        self.R2.grid(row=1, column=0, sticky="nw")
        self.R3 = tk.Radiobutton(self.tab3_left, text="Minor", variable=self.rad_but_var, value=3, command=self.radiobut_chosen)  # lambda:butt3_1.config(state = "normal")
        self.R3.grid(row=2, column=0, sticky="nw")
        self.R4 = tk.Radiobutton(self.tab3_left, text="Aspect", variable=self.rad_but_var, value=4, command=self.radiobut_chosen)  # lambda:butt3_1.config(state = "normal")
        self.R4.grid(row=3, column=0, sticky="nw")
        self.R5 = tk.Radiobutton(self.tab3_left, text="Orientation", variable=self.rad_but_var, value=5, command=self.radiobut_chosen)  # lambda:butt3_1.config(state = "normal")
        self.R5.grid(row=4, column=0, sticky="nw")
        # button to refresh the histogram with adjusted values
        self.butt3_1 = ttk.Button(self.tab3_left, text="Refresh Histogram", command=self.make_hist)
        self.butt3_1["state"] = "disable"
        self.butt3_1.grid(row=7, column=0, sticky='ew')
        # button to plot the ellipses defined by data in histogram
        self.butt3_2 = ttk.Button(self.tab3_left, text="> Plot Ellipses from Histogram", command=self.create_ellipses_image)
        self.butt3_2["state"] = "disable"
        self.butt3_2.grid(row=8, column=0, sticky='ew')
        ttk.Separator(self.tab3_left, orient='horizontal').grid(row=9, column=0, pady=10, sticky="ew")
        # button for crop the underlying dataspace
        self.butt3_3 = ttk.Button(self.tab3_left, text="Redefine Data Borders", command=self.data_list_further)
        self.butt3_3["state"] = "disable"
        self.butt3_3.grid(row=10, column=0, sticky='ew')
        # button for reset dataspace
        self.butt3_4 = ttk.Button(self.tab3_left, text="Reset Data Borders", command=self.reset_data_list)
        self.butt3_4["state"] = "disable"
        self.butt3_4.grid(row=11, column=0, sticky='ew')
        ttk.Separator(self.tab3_left, orient='horizontal').grid(row=12, column=0, pady=10, sticky="ew")
        # button for next tab
        self.butt3_5 = ttk.Button(self.tab3_left, text=">> Final Analysis", command=lambda: self.update_controller_from_tab3(controller))
        self.butt3_5["state"] = "disable"
        self.butt3_5.grid(row=13, column=0, sticky='ew')
        ttk.Separator(master=self.tab3_left, orient='horizontal').grid(row=14, column=0, pady=10, sticky="ew")
        # variables field to insert directl filter variables
        tk.Label(master=self.tab3_left, text='Filter Values').grid(row=15, column=0, sticky='ew')
        self.left_preset_box = tk.Frame(master=self.tab3_left, width=120)
        self.left_preset_box.grid(row=16, column=0, pady=10, sticky="ew")
        self.filterdata = []
        labels_filter = ['Area', 'Major', 'Minor', 'Aspect', 'Orientation']
        for i in range(0, len(labels_filter)):
            tk.Label(self.left_preset_box, text=labels_filter[i]).grid(row=i, column=0)
        self.entryBox = []
        for i in range(0, 10):
            self.entryBox.append(tk.Entry(master=self.left_preset_box, width=10))
            self.entryBox[i].grid(row=i // 2, column=1 + i % 2)
        self.entryBox[0].insert(0, str(self.minimum_area))
        self.entryBox[1].insert(0, str(self.highest_area))
        self.entryBox[2].insert(0, str(self.minimum_major))
        self.entryBox[3].insert(0, str(self.highest_major))
        self.entryBox[4].insert(0, str(self.minimum_minor))
        self.entryBox[5].insert(0, str(self.highest_minor))
        self.entryBox[6].insert(0, str(self.minimum_aspect))
        self.entryBox[7].insert(0, str(self.highest_aspect))
        self.entryBox[8].insert(0, str(self.minimum_orientation))
        self.entryBox[9].insert(0, str(self.highest_orientation))
        for i in range(0, 10):
            self.filterdata.append(float(self.entryBox[i].get()))
        # button to use the values set in the entryBoxes
        self.butt3_6 = ttk.Button(self.tab3_left, text="Use Filters", command=self.create_filtered_particle_list)
        self.butt3_6["state"] = "normal"
        self.butt3_6.grid(row=17, column=0, sticky='ew')
        # button to get the values from the scales into the entryBoxes
        self.butt3_7 = ttk.Button(self.tab3_left, text="Update from Scales", command=self.get_filterdata_from_scale)
        self.butt3_7["state"] = "normal"
        self.butt3_7.grid(row=18, column=0, sticky='ew')

        self.tab3_left.update()

    # displays the histogram to the chosen variable
    def radiobut_chosen(self):
        self.butt3_1.config(state="normal")
        self.scale_for_adjust()
        self.make_hist()
        if self.console_print_on: print("size of tab3_left")
        if self.console_print_on: print(self.tab3_left.winfo_width())

    # create scales for adjusting the histogram
    def scale_for_adjust(self):
        if self.console_print_on: print("A scaling of histogram was done.")
        # set scale boundaries according to choice
        if self.rad_but_var.get() == 1:
            self.scale_minimum = min([P.area for P in self.ParticleProp]) * self.area_factor * 0.99
            self.scale_highest = max([P.area for P in self.ParticleProp]) * self.area_factor * 1.01
        elif self.rad_but_var.get() == 2:
            self.scale_minimum = min([P.major_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 0.99
            self.scale_highest = max([P.major_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 1.01
        elif self.rad_but_var.get() == 3:
            self.scale_minimum = min([P.minor_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 0.99
            self.scale_highest = max([P.minor_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 1.01
        elif self.rad_but_var.get() == 4:
            try:
                self.scale_minimum = min([P.major_axis_length / P.minor_axis_length for P in self.ParticleProp]) * 10 * 0.99
                self.scale_highest = max([P.major_axis_length / P.minor_axis_length for P in self.ParticleProp]) * 10 * 1.01
            except ZeroDivisionError:
                self.scale_minimum = 1
                self.scale_highest = 100  # not higher as aspect 10 possible
                if self.console_print_on: print("ZeroDevisionError")
        elif self.rad_but_var.get() == 5:
            self.scale_minimum = (min([P.orientation for P in self.ParticleProp]) + math.pi / 2) * 180 / math.pi
            self.scale_highest = (max([P.orientation for P in self.ParticleProp]) + math.pi / 2) * 180 / math.pi
        else:
            self.scale_minimum = 0
            self.scale_highest = 100

        self.sc_hist_left = tk.DoubleVar()
        self.sc_hist_rght = tk.DoubleVar()

        self.scale_hist_left = tk.Scale(self.tab3_left, variable=self.sc_hist_left, from_=self.scale_minimum, to=self.scale_highest, length=200, digits=3, orient=tk.HORIZONTAL, resolution=1)
        self.scale_hist_left.set(0)
        self.scale_hist_left.grid(row=5, column=0)

        self.scale_hist_rght = tk.Scale(self.tab3_left, variable=self.sc_hist_rght, from_=self.scale_minimum, to=self.scale_highest, length=200, digits=3, orient=tk.HORIZONTAL, resolution=1)
        self.scale_hist_rght.set(self.scale_highest)
        self.scale_hist_rght.grid(row=6, column=0)  # , sticky="nw")

        if self.console_print_on: print(self.scale_highest)
        if self.console_print_on: print(self.scale_minimum)

    # make the histogram from the variables of the particle/area properties
    def make_hist(self):
        self.butt3_2.config(state="normal")
        # count labels
        MaxLabel = max([P.label for P in self.ParticleProp])
        if self.console_print_on: print(self.rad_but_var.get())
        if self.console_print_on: print("max label " + str(MaxLabel))
        self.hist_list = []
        self.area_no = []

        # prepare the histogram
        self.fig_hist2 = Figure(figsize=(self.img_x, self.img_y))
        tab2_hist = self.fig_hist2.add_subplot(111)

        # choose which parameter for picking and write to hist_list
        if self.rad_but_var.get() == 1:
            tab2_hist.set_title("Area")
            tab2_hist.set_xlabel("Area [nm^2]")
            tab2_hist.set_ylabel("Number Count")
            # make list of areas
            self.minimum_area = float(self.scale_hist_left.get())
            self.highest_area = float(self.scale_hist_rght.get())
            for P in self.ParticleProp:
                if P.area * self.area_factor >= self.minimum_area and P.area * self.area_factor <= self.highest_area:
                    self.hist_list.append(P.area * self.area_factor)
                    self.area_no.append(P.label)

        elif self.rad_but_var.get() == 2:
            tab2_hist.set_title("Major Axis Length")
            tab2_hist.set_xlabel("Major Axis [nm]")
            tab2_hist.set_ylabel("Number Count")
            # make list of major axis
            self.minimum_major = float(self.scale_hist_left.get())
            self.highest_major = float(self.scale_hist_rght.get())
            for P in self.ParticleProp:
                if P.major_axis_length * self.pxl_ratio >= self.minimum_major and P.major_axis_length * self.pxl_ratio <= self.highest_major:
                    self.hist_list.append(P.major_axis_length * self.pxl_ratio)
                    self.area_no.append(P.label)

        elif self.rad_but_var.get() == 3:
            tab2_hist.set_title("Minor Axis Length")
            tab2_hist.set_xlabel("Minor Axis [nm]")
            tab2_hist.set_ylabel("Number Count")
            # make list of minor axis
            self.minimum_minor = float(self.scale_hist_left.get())
            self.highest_minor = float(self.scale_hist_rght.get())
            for P in self.ParticleProp:
                if P.minor_axis_length * self.pxl_ratio >= self.minimum_minor and P.minor_axis_length * self.pxl_ratio <= self.highest_minor:
                    self.hist_list.append(P.minor_axis_length * self.pxl_ratio)
                    self.area_no.append(P.label)

        elif self.rad_but_var.get() == 4:
            tab2_hist.set_title("Aspect Ratio")
            tab2_hist.set_xlabel("Aspect Ratio [/10]")
            tab2_hist.set_ylabel("Number Count")
            # make list of aspects
            self.minimum_aspect = float(self.scale_hist_left.get())
            self.highest_aspect = float(self.scale_hist_rght.get())
            for P in self.ParticleProp:
                if (P.major_axis_length / P.minor_axis_length * 10) >= self.minimum_aspect and (P.major_axis_length / P.minor_axis_length * 10) <= self.highest_aspect:
                    self.hist_list.append(P.major_axis_length / P.minor_axis_length * 10)
                    self.area_no.append(P.label)

        elif self.rad_but_var.get() == 5:
            tab2_hist.set_title("Orientation")
            tab2_hist.set_xlabel("Angle to x-Axis [°]")
            tab2_hist.set_ylabel("Number Count")
            # make list of orientations
            self.minimum_orientation = float(self.scale_hist_left.get())
            self.highest_orientation = float(self.scale_hist_rght.get())
            for P in self.ParticleProp:
                orientation_in_deg = (float(P.orientation) + math.pi / 2) * 180 / math.pi
                if orientation_in_deg >= self.minimum_orientation and orientation_in_deg <= self.highest_orientation:
                    self.hist_list.append(orientation_in_deg)
                    self.area_no.append(P.label)
        else:
            self.hist_list = []

        # make histogram from hist_list
        tab2_hist.hist(self.hist_list)
        canvas3_1 = FigureCanvasTkAgg(self.fig_hist2, self.tab3_right)
        canvas3_1.get_tk_widget().grid(row=0, column=0, rowspan=1)
        # Create vertical lines on the histogram to adjust
        self.lower_limit_line = tab2_hist.axvline(self.scale_hist_left.get(), color='k')
        self.upper_limit_line = tab2_hist.axvline(self.scale_hist_rght.get(), color='k')
        self.scale_hist_left.config(command=self.update_histo_tab3)
        self.scale_hist_rght.config(command=self.update_histo_tab3)
        # toolbar
        toolbar3_1 = NavigationToolbar2Tk(canvas3_1, self.tab3_right, pack_toolbar=False)
        toolbar3_1.update()
        toolbar3_1.grid(row=1, column=0, rowspan=1)

    # when adjust sliders and click button, update histogram
    def update_histo_tab3(self, value):
        # Update the position of the vertical lines
        self.lower_limit_line.set_xdata([self.scale_hist_left.get(), self.scale_hist_left.get()])
        self.upper_limit_line.set_xdata([self.scale_hist_rght.get(), self.scale_hist_rght.get()])
        if self.console_print_on: print("updated hist tab3")
        self.fig_hist2.canvas.draw_idle()

    # make the ellipse image using the region properties
    def create_ellipses_image(self):
        self.butt3_3.config(state="normal")
        self.butt3_5.config(state="normal")
        self.ParticleList = []
        self.ParticleList.append('Partikel_No., posy, posx, minor, major, orientation_in_deg, ratio, area')
        self.ellipse_draw_list = []

        # make image of ellipses on blank image
        # np.zero (blank image) needs dimension of height, width --> using new_imagesize in normal order
        image_zero = np.zeros(shape=[self.new_imagesize[0], self.new_imagesize[1], 3])
        image_zero.fill(0)
        if self.console_print_on: print(image_zero.shape)
        if self.console_print_on: print(self.new_imagesize)

        number = 1
        for P in self.ParticleProp:
            if P.label in self.area_no:
                # get all necessary labled region informations
                major = int(round(P.major_axis_length / 2, 0))
                minor = int(round(P.minor_axis_length / 2, 0))
                posx = int(P.centroid[1])
                posy = int(P.centroid[0])

                # convert orientation to draw ellipses orientation -- since skimage v 0.18 necessary to do pi/2
                orientation = float(P.orientation) + math.pi / 2

                # draw ellipses into blank image at given position in color and do not draw outside of imagesize
                rr, cc = ellipse(posy, posx, minor, major, shape=self.new_imagesize, rotation=orientation)
                try:
                    image_zero[rr, cc] = colors.to_rgb(self.ellipse_color(abs((orientation) / (math.pi))))
                except IndexError:
                    print("one index error with " + str(posy) + "  " + str(posx) + "  " + str(minor) + "  " + str(major))
                    print(rr)  # rr is y value
                    print(cc)  # cc is x value
                org_no = P.label

                # write drawn ellipse data in ellipse_draw_list using image coordinates
                elementi = [number, posy, posx, minor, major, orientation, org_no, cc, rr]
                self.ellipse_draw_list.append(elementi)
                # write properties of drawn data in ParticleList for saving/export
                aspect_ratio = round(P.major_axis_length / P.minor_axis_length, 3)
                ellipse_area = round(math.pi * minor * self.pxl_ratio * major * self.pxl_ratio, 3)
                element = [number, posy, posx, P.minor_axis_length * self.pxl_ratio, P.major_axis_length * self.pxl_ratio, orientation * 180 / math.pi, aspect_ratio, ellipse_area]
                self.ParticleList.append(element)
                # get highest aspect ratio for polarplot
                if aspect_ratio > self.max_aspect:
                    self.max_aspect = aspect_ratio
                number = number + 1
            else:
                if self.console_print_on: print("one data excluded")

        # after drawing every ellipse, safe the image as image_ellipses and number the regions
        self.image_ellipses = image_zero
        self.image_numbers = self.number_region_on_image(self.image_ellipses)
        self.image_ellipses = image_zero

        self.image_ellipses_at_tab3()
        self.make_polarplot()

    # add numbers on the ellipses image to mark the regions
    def number_region_on_image(self, numparray):
        # get an image
        base = Image.fromarray(numparray, mode='RGBA')
        # turn array of new_imagesize
        size = str(self.new_imagesize[1]) + "," + str(self.new_imagesize[0])
        if self.console_print_on: print(base.size)
        if self.console_print_on: print(self.new_imagesize)
        if self.console_print_on: print(size)
        if self.console_print_on: print("here")
        # make a blank image for the text, initialized to transparent text color
        txt = Image.new('RGBA', base.size, (255, 255, 255, 0))
        # get a font
        fnt = ImageFont.truetype('/usr/share/fonts/TTF/arial.ttf', 20)
        # get a drawing context
        d = ImageDraw.Draw(txt)
        # draw text, full opacity - ((x,y),text,font, fill)
        for i in range(len(self.ParticleList) - 1):
            d.text((self.ParticleList[i + 1][2], self.ParticleList[i + 1][1]), str(i + 1), font=fnt, fill=(255, 255, 255, 255))
        return np.array(txt)

    # draw the ellipses on the greyscale image
    def image_ellipses_at_tab3(self):    
        # check if image is already drawn, create if not, update if yes
        if self.ellpsoids_at_tab3_drawn == 0:
            #image ellipses on original image
            self.figd = Figure(figsize=(self.img_x, self.img_y))
            d = self.figd.add_subplot(111)
            self.ellipses_on_img_tab3 = d.imshow(self.image_ellipses, cmap=plt.cm.nipy_spectral)
            d.imshow(self.image, cmap=plt.cm.gray, alpha=0.5)
            self.ellipse_no_img_tab3 = d.imshow(self.image_numbers, cmap=plt.cm.gray, alpha=1)
            if self.image_w_axes:
                d.set_title("Ellipses on Original Image")
                d.set_xlabel("Horizontal Pixel ID")
                d.set_ylabel("Vertical Pixel ID")
            else:
                d.axis("off")
            canvas3_2 = FigureCanvasTkAgg(self.figd, self.tab3_right)
            canvas3_2.get_tk_widget().grid(row=0, column=1, rowspan=1)
            # toolbar
            toolbar3_2 = NavigationToolbar2Tk(canvas3_2, self.tab3_right, pack_toolbar=False)
            toolbar3_2.update()
            toolbar3_2.grid(row=1, column=1, rowspan=1)
            self.ellpsoids_at_tab3_drawn = 1
        else:
            # update image
            self.ellipses_on_img_tab3.set_data(self.image_ellipses)
            self.ellipse_no_img_tab3.set_data(self.image_numbers)
            self.figd.canvas.draw_idle()
            if self.console_print_on: print("ellipses on tab3 updated")

    # create the polarplot
    def make_polarplot(self):
        figc = Figure(figsize=(self.img_x, self.img_y))
        angle = []
        yachse = []
        c = figc.add_subplot(111, projection='polar')
        for i in range(len(self.ParticleList) - 1):
            # orientation in polar coordinates as befor in ellipse_draw_list
            angle.append((self.ParticleList[i + 1][5]) / 180 * math.pi)
            yachse.append(float(self.ParticleList[i + 1][6]))  # aspect ratio
            # mirroring the values to 90°-270° as unidentifiable if 0° or 180°
            angle.append(((self.ParticleList[i + 1][5]) + 180) / 180 * math.pi)
            yachse.append(float(self.ParticleList[i + 1][6]))
        c.set_rmax(self.max_aspect)
        c.plot(angle, yachse, 'bo')
        c.grid(True)
        canvas3_3 = FigureCanvasTkAgg(figc, self.tab3_right)
        canvas3_3.get_tk_widget().grid(row=2, column=0, rowspan=1)
        # toolbar, although zoom does not work with polar plot
        toolbar3_3 = NavigationToolbar2Tk(canvas3_3, self.tab3_right, pack_toolbar=False)
        toolbar3_3.update()
        toolbar3_3.grid(row=3, column=0, rowspan=1)
        self.make_angle_hist()

    # create the angle histogram
    def make_angle_hist(self):
        figE = Figure(figsize=(self.img_x, self.img_y))
        e = figE.add_subplot(111)
        angle = []
        for i in range(len(self.ParticleList) - 1):
            angle.append(self.ParticleList[i + 1][5])
        e.hist(angle, bins=36)  # , rwidth = 5
        if self.image_w_axes:
            e.set_title("Orientation of Ellipses")
            e.set_xlabel("Angle [°]")
            e.set_ylabel("Number Count")
        canvas3_4 = FigureCanvasTkAgg(figE, self.tab3_right)
        canvas3_4.get_tk_widget().grid(row=2, column=1, rowspan=1)
        # toolbar
        toolbar3_4 = NavigationToolbar2Tk(canvas3_4, self.tab3_right, pack_toolbar=False)
        toolbar3_4.update()
        toolbar3_4.grid(row=3, column=1, rowspan=1)

    # crop the datalist when hit button 'Redefine Data Borders' or 'Use Filters'
    def data_list_further(self):
        self.second_datalist = []
        if self.reset_ParticleProp_list is None:
            if self.console_print_on: print(locals())
            self.reset_ParticleProp_list = []
            for P in self.ParticleProp:
                self.reset_ParticleProp_list.append(P)
            self.data_excluded = 0
        for P in self.ParticleProp:
            if P.label in self.area_no:
                self.second_datalist.append(P)
            else:
                self.data_excluded = self.data_excluded + 1
        if self.console_print_on: print("ParticleProp: " + str(len(self.ParticleProp)))
        if self.console_print_on: print("reset List: " + str(len(self.reset_ParticleProp_list)))
        if self.console_print_on: print("Excluded: " + str(self.data_excluded))
        Length_list2 = len(self.second_datalist)
        if self.console_print_on: print("New ParticleProp: " + str(Length_list2))
        self.ParticleProp = self.second_datalist
        self.butt3_4.config(state="normal")
        self.scale_for_adjust()
        self.make_hist()

    # reset the cropped datalist to the list handed over by tab 2
    def reset_data_list(self):
        self.ParticleProp = self.reset_ParticleProp_list
        self.butt3_4.config(state="disabled")
        self.scale_for_adjust()
        self.make_hist()

    # filter the particle list of tab 2 according to the entryBox values
    def create_filtered_particle_list(self):
        self.area_no = []
        self.filterdata = []
        for i in range(0, 10):
            self.filterdata.append(float(self.entryBox[i].get()))
        if self.console_print_on: print("The following filter data was used: " + str(self.filterdata))
        for P in self.ParticleProp:
            orientation_in_deg = (float(P.orientation) + math.pi / 2) * 180 / math.pi
            if P.area * self.area_factor >= self.filterdata[0] and P.area * self.area_factor <= self.filterdata[1]:
                if P.major_axis_length * self.pxl_ratio >= self.filterdata[2] and P.major_axis_length * self.pxl_ratio <= self.filterdata[3]:
                    if P.minor_axis_length * self.pxl_ratio >= self.filterdata[4] and P.minor_axis_length * self.pxl_ratio <= self.filterdata[5]:
                        if P.major_axis_length / P.minor_axis_length * 10 >= self.filterdata[6] and P.major_axis_length / P.minor_axis_length * 10 <= self.filterdata[7]:
                            if orientation_in_deg >= self.filterdata[8] and orientation_in_deg <= self.filterdata[9]:
                                self.area_no.append(P.label)
        # write filter in list ['area', 'major', 'minor', 'aspect', 'orientation']
        self.minimum_area = self.filterdata[0]
        self.highest_area = self.filterdata[1]
        self.minimum_major = self.filterdata[2]
        self.highest_major = self.filterdata[3]
        self.minimum_minor = self.filterdata[4]
        self.highest_minor = self.filterdata[5]
        self.minimum_aspect = self.filterdata[6]
        self.highest_aspect = self.filterdata[7]
        self.minimum_orientation = self.filterdata[8]
        self.highest_orientation = self.filterdata[9]
        self.rad_but_var.set(1)
        self.butt3_1.config(state="normal")
        self.data_list_further()
        self.create_ellipses_image()

    # update the entryBoxes with the current datalist values (rounded to integer)
    def get_filterdata_from_scale(self):
        self.update_min_max()
        # write filter in list ['area', 'major', 'minor', 'aspect', 'orientation']
        self.filterdata[0] = self.minimum_area
        self.filterdata[1] = self.highest_area
        self.filterdata[2] = self.minimum_major
        self.filterdata[3] = self.highest_major
        self.filterdata[4] = self.minimum_minor
        self.filterdata[5] = self.highest_minor
        self.filterdata[6] = self.minimum_aspect
        self.filterdata[7] = self.highest_aspect
        self.filterdata[8] = self.minimum_orientation
        self.filterdata[9] = self.highest_orientation
        # update entry fields
        for i in range(0, 10):
            self.entryBox[i].delete(0, 'end')
            self.entryBox[i].insert(0, str(self.filterdata[i]))

    # update the variables min/max values of the current particle/area list
    def update_min_max(self):
        self.minimum_area = round(min([P.area for P in self.ParticleProp]) * self.area_factor * 0.99, 0)
        self.highest_area = round(max([P.area for P in self.ParticleProp]) * self.area_factor * 0.99 + 1, 0)
        self.minimum_major = round(min([P.major_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 1.01)
        self.highest_major = round(max([P.major_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 1.01 + 1, 0)
        self.minimum_minor = round(min([P.minor_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 0.99)
        self.highest_minor = round(max([P.minor_axis_length for P in self.ParticleProp]) * self.pxl_ratio * 0.99 + 1, 0)
        try:
            self.minimum_aspect = round(min([P.major_axis_length / P.minor_axis_length for P in self.ParticleProp]) * 10 * 0.99)
            self.highest_aspect = round(max([P.major_axis_length / P.minor_axis_length for P in self.ParticleProp]) * 10 * 0.99 + 1, 0)
        except ZeroDivisionError:
                self.minimum_aspect = 1
                self.highest_aspect = 100
        self.minimum_orientation = round((min([P.orientation for P in self.ParticleProp]) + math.pi / 2) * 180 / math.pi)
        self.highest_orientation = round((max([P.orientation for P in self.ParticleProp]) + math.pi / 2) * 180 / math.pi + 1, 0)

    # update images and then handover the altered parameters to the controller, so they can be used with next tab
    def update_controller_from_tab3(self, controller):
        self.get_filterdata_from_scale()
        # variables to controller
        controller.ellipse_draw_list = self.ellipse_draw_list
        controller.ParticleList = self.ParticleList
        controller.ParticleProp = self.ParticleProp
        controller.highest_major = self.highest_major
        controller.minimum_major = self.minimum_major
        controller.image_ellipses = self.image_ellipses
        controller.image_numbers = self.image_numbers
        controller.filter_used_list = self.filterdata
        # next tab 4
        controller.next_step(4)
        controller.show_frame(3)

    # handover when saving the parameters
    def make_save_update_from_tab3(self, controller):
        #variables to controller
        controller.ellipse_draw_list = self.ellipse_draw_list
        controller.filter_used_list = self.filterdata
