#####################################
# ################################# #
# ### Tab 4 - Ellipse picking - ### #
# ################################# #
#####################################

# import packages
import matplotlib
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.backends.backend_tkagg import NavigationToolbar2Tk
from matplotlib.figure import Figure
from matplotlib.pyplot import rcParams
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import numpy as np
import tkinter as tk
from tkinter import ttk
import os
import datetime
import math
import pandas as pd

# global plot parameters
matplotlib.use("TkAgg")
LARGE_FONT = ("Verdana", 12)
GlobFontSize = 10
params = {
    'axes.labelsize': GlobFontSize,
    'font.size': GlobFontSize,
    'axes.titlesize': GlobFontSize,
    'font.family': "Calibri",
    'legend.fontsize': GlobFontSize,
    'xtick.labelsize': GlobFontSize,
    'ytick.labelsize': GlobFontSize,
    'text.usetex': False,
    'figure.constrained_layout.use': True,
    'lines.markersize': 3.0,
    'boxplot.flierprops.markersize': 3
}
rcParams.update(params)


# tab 4 code
class Page_EllipsePicking(tk.Frame):
    def __init__(self, parent, controller):
        # get dummys and variables from controller
        self.console_print_on = controller.console_print_on
        if self.console_print_on: print("statistic Page")
        self.ParticleList = controller.ParticleList
        self.ParticleProp = controller.ParticleProp
        self.image_path = controller.image_path
        self.img_x = controller.img_x
        self.img_y = controller.img_y
        self.ellipse_draw_list = controller.ellipse_draw_list
        self.new_imagesize = controller.new_imagesize
        self.image_w_axes = controller.image_w_axes.get()
        self.image = controller.image
        self.image_numbers = controller.image_numbers
        self.image_rgb = controller.image_rgb
        self.image_ellipses = controller.image_ellipses
        self.pxl_ratio = controller.pxl_ratio
        self.particles_at_border = controller.particles_at_border
        self.analyse_orientations_img = controller.analyse_orientations_img

        # variables of the tab
        self.was_not_clicked = True
        self.image_ellipsoids_at_tab4_drawn = 0
        self.var_feld = []
        # copy ellipse_draw_list to workingList
        self.DeleteList = []
        self.WorkingList = []
        for e in self.ellipse_draw_list:
            self.WorkingList.append(e)
        self.final_ParticleList = []
        self.old_ParticleList = []
        # create final images
        self.finalfigure = Figure(figsize=(self.img_x, self.img_y))
        if self.console_print_on: print(self.new_imagesize)
        # attention: figure needs (width, height), new_imagesize is (height, width)
        # --> need to be turned
        self.finalfig_orgdim = Figure(figsize=(self.new_imagesize[1] / 100, self.new_imagesize[0] / 100))

        ## create tab layout ##
        tk.Frame.__init__(self, parent)
        # devide window into 2 frames
        self.tab4_left = ttk.Frame(master=self)
        self.tab4_left.grid(row=0, column=0, sticky="nw")
        self.tab4_left.columnconfigure(0, minsize=220)
        self.tab4_right = ttk.Frame(master=self)
        self.tab4_right.grid(row=0, column=1, sticky="nsew")

        # make tab 4 menu on left container
        # toggle numbers on image
        self.number_show = tk.BooleanVar()
        self.number_show.set(True)
        self.check_numbers = ttk.Checkbutton(self.tab4_left, text="Switch ON Numbers", variable=self.number_show, command=self.image_ellipsoids_at_tab4)
        self.check_numbers.grid(row=0, column=0, sticky='ew')
        # button for save/export data
        self.butt4_3 = ttk.Button(self.tab4_left, text=">> Save/Export Data", command=lambda: controller.write_datafile())
        self.butt4_3.grid(row=2, column=0, sticky='ew', pady=5)
        ttk.Separator(self.tab4_left, orient='horizontal').grid(row=3, column=0, pady=10, sticky="ew")
        # button for the orientation plot of the visible image
        self.butt4_1 = ttk.Button(self.tab4_left, text="Plot Orientations", command=self.plot_orientations, state="enabled")
        self.butt4_1.grid(row=4, column=0, sticky='ew')
        # button for a distance plot
        self.butt4_4 = ttk.Button(self.tab4_left, text="Plot Distances", command=lambda: self.plot_distances(controller), state="enabled")
        self.butt4_4.grid(row=5, column=0, sticky='ew')
        # button for the angle analysis
        self.butt4_6 = ttk.Button(self.tab4_left, text="Analyse Angles", command=self.analyse_orientations)
        self.butt4_6.grid(row=7, column=0, sticky='ew')
        ttk.Separator(self.tab4_left, orient='horizontal').grid(row=8, column=0, pady=10, sticky="ew")
        # button for clearing all lables on image
        self.butt4_7 = ttk.Button(self.tab4_left, text="Clear Labels on Image", command=self.Button_deleteAllonImage, state="enabled")
        self.butt4_7.grid(row=9, column=0, sticky='ew')

        # start final evaluation
        self.finall_ellipsoids_image()
        self.create_final_ParticleList()

    # create window displaying the orientation plot
    def plot_orientations(self):
        angle = []
        yaxe = []
        fig_pol = Figure(figsize=(self.img_x, self.img_y))
        if self.console_print_on: print(self.img_x)
        if self.console_print_on: print(self.img_y)
        if self.console_print_on: print(fig_pol.dpi)
        if self.console_print_on: print("done")
        c = fig_pol.add_subplot(121, projection='polar')
        for i in range(len(self.final_ParticleList) - 1):
            # orientation in polar coordinates as befor in ellipse_draw_list
            angle.append((self.final_ParticleList[i + 1][5]) / 180 * math.pi)
            yaxe.append(float(self.final_ParticleList[i + 1][6]))  # aspect ratio
            # mirroring the values to 90°-270° as unidentifiable if 0° or 180°
            angle.append(((self.final_ParticleList[i + 1][5]) + 180) / 180 * math.pi)
            yaxe.append(float(self.final_ParticleList[i + 1][6]))
        c.plot(angle, yaxe, 'b.')
        c.grid(True)
        c.set_rmax(np.max(yaxe) + 0.5)
        c.set_title('Orientation of Ellipses on Image')
        c.set_rlabel_position(-22.5)  # get radial labels away from plotted line

        # add a histogram
        angle_deg = []
        for i in angle:
            angle_deg.append(i * 180 / math.pi)
        h_orient = fig_pol.add_subplot(122)
        h_orient.hist(angle_deg, bins=120, range=(0, 360))
        h_orient.set_xlabel("Angle [°]")
        h_orient.set_ylabel("Number Count")
        h_orient.set_xlim(0, 360)
        h_orient.set_title("Angle Histogram")

        self.window_pol = tk.Tk()
        self.window_pol.geometry(str(int(self.img_x * fig_pol.dpi)) + "x" + str(int(self.img_y * fig_pol.dpi) + 50))
        canvas_polar = FigureCanvasTkAgg(fig_pol, self.window_pol)
        canvas_polar.get_tk_widget().grid(row=0, column=0, rowspan=1)
        # toolbar
        toolbar_polar = NavigationToolbar2Tk(canvas_polar, self.window_pol, pack_toolbar=False)
        toolbar_polar.update()
        toolbar_polar.grid(row=1, column=0, rowspan=1)
        self.final_polar_plot=fig_pol

    # clear the lables on the image
    def Button_deleteAllonImage(self):
        self.DeleteList = []
        for P in self.ellipse_draw_list:
            self.DeleteList.append(P[6])
        self.WorkingList = []
        self.was_not_clicked = False
        self.finall_ellipsoids_image()
        self.create_final_ParticleList()

    # event handler when clicked on image
    def onclick(self, event):
        if self.was_not_clicked:
            self.DeleteList = []
            self.WorkingList = []
            for e in self.ellipse_draw_list:
                self.WorkingList.append(e)
            if self.console_print_on: print(len(self.ellipse_draw_list))
            self.was_not_clicked = False
        # when there was a click on the image area
        if not event.xdata == None and not event.ydata == None:
            if self.console_print_on:
                print('%s click: button = %d, x = %d, y = %d, xdata = %f, ydata = %f' %
                ('double' if event.dblclick else 'single', event.button,
                event.x, event.y, round(event.xdata, 0), round(event.ydata, 0)))

            # ellipse draw list element [number, posy, posx, minor, major, orientation, org_no, array(cc), array(rr)]
            for i in self.ellipse_draw_list:
                if [int(round(event.xdata, 0)), int(round(event.ydata, 0))] in list(map(list, zip(i[7], i[8]))):
                    if self.console_print_on: print("clicked on particle region")
                    # click is on Particle -> delete from workingList
                    if not i[6] in self.DeleteList:
                        self.DeleteList.append(i[6])
                        for el in self.WorkingList:
                            if i[6] == el[6]:
                                self.WorkingList.remove(el)
                                if self.console_print_on: print("one particle deleted from WorkingList")
                            else:
                                if self.console_print_on: print("is not particle to be removed")
                    # Particle was formerly deleted -> restore
                    else:
                        self.WorkingList.append(i)
                        for eli in self.DeleteList:
                            if i[6] == eli:
                                if self.console_print_on: print("one particle deleted from DeleteList")
                                self.DeleteList.remove(eli)
                        if self.console_print_on: print("Particle restored")                        

            if self.console_print_on: print(self.DeleteList)
            # update image and list
            self.finall_ellipsoids_image()
            self.create_final_ParticleList()

        # when click was not on image area
        else:
            if self.console_print_on: print("out of region click")

    # update particle list
    def create_final_ParticleList(self):
        if self.console_print_on: print("start final particle list")
        self.old_ParticleList = self.ParticleList
        new_list = []
        new_list.append(self.ParticleList[0])
        # working list has draw infos but not actual properties, so that compare with original particle list is necessary
        for x in range(0, len(self.WorkingList), 1):
            for k in range(1, len(self.ParticleList), 1):
                if self.WorkingList[x][0] == self.ParticleList[k][0]:
                    new_list.append(self.ParticleList[k])
        self.final_ParticleList = new_list
        if self.console_print_on: print("new_list ready")

    # create image of colorful ellipses with given coordinates in working list
    # WorkingList with elements [number, posy, posx, minor, major, orientation, org_no, array(cc), array(rr)]
    def finall_ellipsoids_image(self):
        self.ellipse_color = plt.cm.hsv
        # create black image to draw ellipses -> np.zeros need shape like new_imagesize (height, width)
        image_zero = np.zeros([self.new_imagesize[0], self.new_imagesize[1], 3])
        image_zero.fill(0)
        # draw colorful ellipses from workinglist with the cc,rr coordinate arrays at
        # position 7 (x-coord) and 8 (y-coord) - note that y,x coords from np array
        for i in range(len(self.WorkingList)):
            image_zero[self.WorkingList[i][8], self.WorkingList[i][7]] = colors.to_rgb(self.ellipse_color(abs((self.WorkingList[i][5]) / (math.pi))))
        self.image_ellipses = image_zero
        self.image_ellipsoids_at_tab4()

    # image ellipsoids on original image with or without numbers
    def image_ellipsoids_at_tab4(self):
        # if no image is drawn yet
        if self.image_ellipsoids_at_tab4_drawn == 0:
            self.fig_el = Figure(figsize=(self.img_x * 2, self.img_y * 2 + 0.4))
            ellipse_img = self.fig_el.add_subplot(111)
            self.ellipsoids_on_img_tab4 = ellipse_img.imshow(self.image_ellipses, cmap=plt.cm.nipy_spectral)
            ellipse_img.imshow(self.image_rgb, cmap=plt.cm.gray, alpha=0.6)
            # draw numbers if checkbox is ON
            if self.number_show.get():
                self.number_show_img = ellipse_img.imshow(self.image_numbers, cmap=plt.cm.gray, alpha=1)
            if self.image_w_axes:
                ellipse_img.axis("on")
                ellipse_img.set_title("Ellipses on Original Image")
                ellipse_img.set_xlabel("Horizontal Pixel ID")
                ellipse_img.set_ylabel("Vertical Pixel ID")
            else:
                ellipse_img.axis("off")

            canvas4 = FigureCanvasTkAgg(self.fig_el, self.tab4_right)
            # create mouse click event to choose ellipses
            canvas4.mpl_connect('button_press_event', self.onclick)
            canvas4.get_tk_widget().grid(row=0, column=0, sticky="nsew", columnspan=1)
            self.image_ellipsoids_at_tab4_drawn = 1
        # if image is present -> update
        else:
            # update image
            self.ellipsoids_on_img_tab4.set_data(self.image_ellipses)
            if self.number_show.get() == False:
                self.number_show_img.set_alpha(0)
            else:
                self.number_show_img.set_alpha(1)
            self.fig_el.canvas.draw_idle()
            if self.console_print_on: print("ellipsoids on tab3 updated")

        # adress image on display for later save
        self.finalfigure = self.fig_el

        # create image with orginial dimensions, for saving later on
        # attention: figure needs (width, height), new_imagesize is (height, width) --> need to be turned
        fig_el_orgdim = Figure(figsize=(self.new_imagesize[1] / 100, self.new_imagesize[0] / 100))
        ellipse_img_orgdim = fig_el_orgdim.add_subplot(111)
        ellipse_img_orgdim.imshow(self.image_ellipses, cmap=plt.cm.nipy_spectral)
        ellipse_img_orgdim.imshow(self.image_rgb, cmap=plt.cm.gray, alpha=0.6)
        ellipse_img_orgdim.axis("off")
        self.finalfig_orgdim = fig_el_orgdim
        if self.console_print_on: print('the dimensions of image prior save: {}'.format(self.finalfig_orgdim.get_size_inches()))
        # set (width, height), but from shape get (height, width), do use original pixel size, else devide 100
        self.finalfig_orgdim.set_size_inches(self.new_imagesize[1], self.new_imagesize[0])

    # update images and then handover the altered parameters to the controller
    def update_controller_from_tab4(self, controller):
        self.plot_orientations()
        controller.final_ParticleList = self.final_ParticleList
        controller.ParticleList = self.ParticleList
        controller.finalfigure = self.finalfigure
        controller.finalfig_orgdim = self.finalfig_orgdim
        controller.image_ellipses = self.image_ellipses
        controller.final_polar_plot=self.final_polar_plot
        controller.analyse_orientations_img = self.analyse_orientations_img
        if self.console_print_on: print("Controller updated")

    # handover when saving the parameters
    def make_save_update_from_tab4(self, controller):
        if controller.save_angle_dist.get(): self.plot_orientations()
        controller.final_ParticleList = self.final_ParticleList
        controller.finalfigure = self.finalfigure
        if controller.save_angle_dist.get(): controller.final_polar_plot=self.final_polar_plot        
        controller.finalfig_orgdim = self.finalfig_orgdim
        controller.image_ellipses = self.image_ellipses
        controller.analyse_orientations_img = self.analyse_orientations_img

###########################
## ##################### ##
## # Statistic windows # ##
## ##################### ##
###########################
    # calculate distances from particles
    def create_distance_list(self, particles):
        dist_list = []
        p = particles
        # create list: x coordinate (in particleList[element][0]), y, distance, angle
        # (not used) input-particle list style: "Partikel_No.","posy", "posx", "minor", "major", "orientation_in_deg", "ratio"  
        # input with working list [number, posy, posx, minor, major, orientation, org_no]
        for P1 in range(0,len(p),1):
            for P2 in range(P1+1,len(p),1):
                d = math.sqrt((p[P2][2]-p[P1][2])**2 + (p[P2][1]-p[P1][1])**2) * self.pxl_ratio # in nm
                try:
                    deg = math.degrees(-math.atan((p[P2][1] - p[P1][1]) / (p[P2][2] - p[P1][2])))
                except:
                    deg = 90

                element = [p[P1][0], p[P2][0], d, deg]   # orientation in polar coordinates as befor in ellipse_draw_list
                dist_list.append(element)
                element2 = [p[P1][0],p[P2][0],d, deg + 180]   # double orientation in polar coordinates as befor in ellipse_draw_list
                dist_list.append(element2)
        if self.console_print_on: print("distance list created")
        return dist_list

    # show the distances of the particles
    def plot_distances(self, controller):
        particle_distance = self.create_distance_list(self.WorkingList)
        xdist_achse = []
        angle_dist_achse = []
        for i in range(len(particle_distance)):
            # write distances into drawlist
            xdist_achse.append(particle_distance[i][2])
            # write angles into draw list
            angle_dist_achse.append((particle_distance[i][3]) / 180 * math.pi)
        # plot histogram and polarplot
        self.window_distances = tk.Tk()
        self.window_distances.geometry("1000x800")
        figt = Figure(figsize=(self.img_x, self.img_y))
        be = figt.add_subplot(121)
        be.hist(xdist_achse, bins=100)
        be.set_title("Distance")
        be.set_xlabel("Distance [nm]")
        be.set_ylabel("Numbers")
        de = figt.add_subplot(122, projection='polar')
        de.plot(angle_dist_achse, xdist_achse, marker=".", linestyle="None", markersize=1)
        canvas_polar2 = FigureCanvasTkAgg(figt, self.window_distances)
        canvas_polar2.get_tk_widget().grid(row=0, column=1, rowspan=1)
        # toolbar
        toolbar_polar2 = NavigationToolbar2Tk(canvas_polar2, self.window_distances, pack_toolbar=False)
        toolbar_polar2.update()
        toolbar_polar2.grid(row=1, column=1, rowspan=1)
        # save the distance figure imediately to folder
        time = datetime.datetime.now().strftime("%Y-%m-%d %H-%M-%S")
        split_name = os.path.splitext(str(controller.image_path[1]))
        figt.savefig((str(controller.image_path[0]) + "\\" + str(split_name[0]) + '_distancePlot_' + time + ".png"), dpi=100, bbox_inches="tight", pad_inches=0)

    # orientation correlation
    def analyse_orientations(self):
        dist_list = []
        p = self.final_ParticleList[1:]
        # create list: x coordinate (in particleList[element][0]), y, distance, angle
        # (not used) input-particle list style: "Partikel_No.","posy", "posx", "minor", "major", "orientation_in_deg", "ratio"  
        # input with working list [number, posy, posx, minor, major, orientation, org_no]
        for P1 in range(0,len(p),1):
            for P2 in range(P1+1,len(p),1):
                d = math.sqrt((p[P2][2]-p[P1][2])**2 + (p[P2][1]-p[P1][1])**2)
                if p[P1][5]>p[P2][5]:
                    compare_angle = round((p[P1][5]-p[P2][5]), 0)
                else:
                    compare_angle = round((p[P2][5]-p[P1][5]), 0)
                # print('angle 1 '+str(p[P1][5])+ '  angle 2 '+str(p[P2][5]))
                # the angles inbetween the ellipses vectors are independent of direction, just angles of 0-90° are possible
                if compare_angle>180:
                    compare_angle = 360 - compare_angle
                if compare_angle>90 and compare_angle<=180:
                    compare_angle = 180 - compare_angle
                try:
                    deg = math.degrees(-math.atan((p[P2][1] - p[P1][1]) / (p[P2][2] - p[P1][2])))
                except:
                    deg = 90

                element = [p[P1][0], p[P2][0], d, deg, compare_angle]   # orientation in polar coordinates as befor in ellipse_draw_list
                dist_list.append(element)

        xdist_achse = []
        compare_angle_dist = []
        for i in range(len(dist_list)):
            # write distances into drawlist
            xdist_achse.append(dist_list[i][2])
            # write angles into drawlist
            compare_angle_dist.append(dist_list[i][4])

        dataframe = pd.DataFrame(xdist_achse, columns=['distance_pxl'])
        dataframe['compare_angle'] = compare_angle_dist
        binsize = 10
        dataframe['bins'] = dataframe['distance_pxl'] // binsize
        mean_dataframe = dataframe.groupby('bins').mean()
        # create figure        
        analyse_orientations_img = Figure(figsize=(3, 2.2))
        ada = analyse_orientations_img.add_subplot()
        if len(dataframe.index) < 10:
            ada.plot(mean_dataframe['distance_pxl'] * self.pxl_ratio, np.cos(2 / 180 * np.pi * mean_dataframe['compare_angle']), 'ro', label='mean')
        else:
            ada.plot(mean_dataframe['distance_pxl'] * self.pxl_ratio, np.cos(2 / 180 * np.pi * mean_dataframe['compare_angle']), 'r', label='mean')
        ada.set_title('Evaluation of Angles')
        ada.set_ylabel('Mean Angle cos(2*({0}(0)-{0}(r)))'.format(chr(952)))
        ada.set_xlabel('Distance r [nm]')
        ada.set_ylim(-1.05, 1.05)
        self.analyse_orientations_img = analyse_orientations_img
        # show figure
        self.window_poly = tk.Tk()
        canvas_ada = FigureCanvasTkAgg(analyse_orientations_img, self.window_poly)
        canvas_ada.get_tk_widget().grid(row=0)
        # toolbar
        toolbar_polar = NavigationToolbar2Tk(canvas_ada, self.window_poly, pack_toolbar=False)
        toolbar_polar.update()
        toolbar_polar.grid(row=1)
